/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Marcell Gal cell at x-dsl.hu]                *
***********************************************************
* used for some simple arithmetics (included, not linked) *
\*********************************************************/

#ifndef __ARITHM_C__
#define __ARITHM_C__

   /*!
   * \ingroup g_math
   *
   * \brief 16x16 Multiply.
   *
   * details ...
   *
   * \return unsigned 32 bit value
   */
static inline uint32_t mult16_16(uint16_t a, uint16_t b) {
	uint32_t result;
	uint8_t tmp;
/*
%A0 to refer to the lowest byte of the first operand,
%A1 to the lowest byte of the second operand and so on...

%n0 : result
%n1 : tmp
%n2 : a
%n3 : b
*/

	asm (
		"\n"
		"clr %1 \n\t"
		"mul %B2, %B3 \n\t"  // hi(a) * hi(b)
		"movw %C0, r0 \n\t"
		"mul %A2, %A3 \n\t"  // lo(a) * lo(b)
		"movw %A0, r0 \n\t"
		"mul %B2, %A3 \n\t"  // hi(a) * lo(b)
		"add %B0, r0 \n\t"
		"adc %C0, r1 \n\t"
		"adc %D0, %1 \n\t"
		"mul %A2, %B3 \n\t"  // lo(a) * hi(b)
		"add %B0, r0 \n\t"
		"adc %C0, r1 \n\t"
		"adc %D0, %1 \n\t"
		"clr r1 \n\t"
		: "=&r" (result), "=&r" (tmp)
		: "r" (a), "r" (b)
	);

	return result;
}

   /*!
   * \ingroup g_math
   *
   * \brief 16x8 Multiply.
   *
   * details ...
   *
   * \return unsigned 32 bit value
   */
static inline uint32_t mult16_8(uint16_t a, uint8_t b) {
	uint32_t result;

/*
%A0 to refer to the lowest byte of the first operand,
%A1 to the lowest byte of the second operand and so on...

%n0 : result
%n1 : a
%n2 : b
*/

	asm (
		"\n"
		"clr %D0 \n\t"
		"clr %C0 \n\t"
		"mul %A1, %A2 \n\t"  // lo(a) * lo(b)
		"movw %A0, r0 \n\t"
		"mul %B1, %A2 \n\t"  // hi(a) * lo(b)
		"add %B0, r0 \n\t"
		"adc %C0, r1 \n\t"	// note that D0 stays 0
		"clr r1 \n\t"
		: "=&r" (result)
		: "r" (a), "r" (b)
	);

	return result;
}

   /*!
   * \ingroup g_math
   *
   * \brief Logical shift right 24 bits.
   * \return unsigned 8 bit value
   */
static inline uint8_t lsr24(uint32_t a)
{
	uint8_t result;
	asm(
		"\n"
		"mov %A0, %D1 \n\t"
		: "=&r" (result)
		: "r" (a)
	);
	return result;
}

   /*!
   * \ingroup g_math
   *
   * \brief Logical shift right 16 bits.
   *
   * \return unsigned 8 bit value
   */
static inline uint8_t lsr16(uint32_t a)
{
	uint8_t result;
	asm(
		"\n"
		"mov %A0, %C1 \n\t"
		: "=&r" (result)
		: "r" (a)
	);
	return result;
}

/* just to avoid the 12*cycle (full 4byte each) gcc generates
*/
   /*!
   * \ingroup g_math
   *
   * \brief 16 bit divide.
   *
   * just to avoid the 12*cycle (full 4byte each) gcc generates
   *
   * \return unsigned 16 bit value
   */
static inline uint16_t div4096(uint32_t a) {
	uint16_t result;

/*
%A0 to refer to the lowest byte of the first operand,
%A1 to the lowest byte of the second operand and so on...

%n0 : result
%n1 : a
*/
	asm(
		"\n"
#if 1
		"swap %B1 \n\t"
		"swap %C1 \n\t"
		"swap %D1 \n\t"
		"ldi %A1, 0xf0 \n\t"

		"eor %D1, %C1 \n\t"
		"and %D1, %A1 \n\t"
		"eor %D1, %C1 \n\t"

		"eor %C1, %B1 \n\t"
		"and %C1, %A1 \n\t"
		"eor %C1, %B1 \n\t"
		"movw %A0, %C1 \n\t"
		: "=&r" (result)
		: "d" (a)	// 'r' is r[0..31], 'd' is r[16..31]
#else
		"movw %A0, %C1 \n\t"

		"lsl %B1 \n\t"
		"rol %A0 \n\t"
		"rol %B0 \n\t"

		"lsl %B1 \n\t"
		"rol %A0 \n\t"
		"rol %B0 \n\t"

		"lsl %B1 \n\t"
		"rol %A0 \n\t"
		"rol %B0 \n\t"

		"lsl %B1 \n\t"
		"rol %A0 \n\t"
		"rol %B0 \n\t"
		: "=&r" (result)
		: "r" (a)
#endif
	);
	return result;
}

/* signed multiplication with offset trick
a*b / 256
*/
   /*!
   * \ingroup g_math
   *
   * \brief Multiply 8 bit values with divide by 265
   *
   * details ...
   *
   * \return signed 8 bit value
   */
static inline int8_t mult8i_8_div256(int8_t a, uint8_t b) {
	uint8_t t;
	t = (uint8_t)a + 0x80; // we offset it to avoid sign-problems
	t = (t * b) >> 8;
	return t - (b >> 1); // we revert the +0x80 sign-trick
}

/* the 0x80 could be a parameter, since it's sometimes used as 100 (eg.: adjust) */
   /*!
   * \ingroup g_math
   *
   * \brief Multiply 8 bit values with offset
   *
   * details ...
   *
   * \return signed 16 bit value
   */
static inline int16_t mult8i_8(int8_t a, uint8_t b) {
	uint8_t t;
	uint16_t t16;
	t = (uint8_t)a + 0x80; // we offset it to avoid sign-problems
	t16 = t * b;
	return t16 - (uint16_t)(b * 0x80); // we revert the +0x80 sign-trick
}

/* returns a * b */
   /*!
   * \ingroup g_math
   *
   * \brief Multiply 16 bit values
   *
   * details ...
   *
   * \return signed 32 bit value
   */
static inline int32_t mult16i_16(int16_t a, uint16_t b) {
	uint16_t t;
	uint32_t k;
#if 0
	t = (uint16_t)a + 0x8000; // we offset it to avoid sign-problems
	k = mult16_16(t, b);
	return k - mult16_16(b, 0x8000); // we revert the +0x8000 sign-trick
#else
	uint8_t sign;
	sign = a>>8;
	t = ((sign & 0x80) ? 0-a : a);
	k = mult16_16(t, b);
	return ((sign & 0x80) ? 0-k : k);
#endif
}

   /*!
   * \ingroup g_math
   *
   * \brief Square 16 bit values
   *
   * details ...
   *
   * \return unsigned 32 bit value
   */
static inline uint32_t square_16i(int16_t a) {
	uint16_t t;
	uint8_t sign;

	sign = a>>8;
	t = ((sign & 0x80) ? 0-a : a);
	return mult16_16(t, t);
}

/* returns a * b / 256 */
   /*!
   * \ingroup g_math
   *
   * \brief Multiply 16 bit value by 8 bit value / 256
   *
   * details ...
   *
   * \return unsigned 16 bit value : a * b / 256
   */
static inline int16_t mult16i_8_div256(int16_t a, uint8_t b) {
	uint16_t t;
	t = (uint16_t)a + 0x8000; // we offset it to avoid sign-problems
	t = mult16_8(t, b) >> 8;
	return t - (uint16_t)(b * 0x80); // we revert the +0x80 sign-trick
}

/* returns a * b / 256 */
   /*!
   * \ingroup g_math
   *
   * \brief Multiply 16 bit value by 8 bit value / 256  (signed)
   *
   * details ...
   *
   * \return signed 16 bit value : a * b / 256
   */
static inline int32_t mult16i_16_div256(int16_t a, uint16_t b) {
	uint16_t t;
	uint32_t k;
	t = (uint16_t)a + 0x8000; // we offset it to avoid sign-problems
	k = mult16_16(t, b) >> 8;
	return k - mult16_8(b, 0x80); // we revert the +0x8000 sign-trick
}


/* signed downscale (divide) by 2th exponent */
   /*!
   * \ingroup g_math
   *
   * \brief Divide 16 bit value by 8 bit exponent
   *
   * signed downscale (divide) by 2th exponent
   *
   * \return a >> exp
   */
static inline int16_t div16i_exp2(int16_t a, uint8_t exp) {
	/* for given exponent it could be tricked to something like (but not more efficient :-( ):
	if(exp == 4){
		uint16_t t;
		uint8_t sign;
		t = a;
		sign = t >> 8;
		t >>= exp; // this is unsigned, so GCC can use a swap
		if(sign & 0x80) // aha, it was negative!
			return t + 0xF000;

		return t;
	}
	*/
	// but generally GCC does a good job with (asr;ror) exp times
	return a >> exp;
}

/* scalar multiple of 2x2 matrices:
element by element unsigned multiplication and
accumulate. Used in the table_lookups.
t. are 8 bit table-data and w. are 16 bit weights
*/
   /*!
   * \ingroup g_math
   *
   * \brief Scaler Multiply of 2x2 Matrices
   *
   * scalar multiple of 2x2 matrices:
   * element by element unsigned multiplication and
   * accumulate. Used in the table_lookups.
   * t. are 8 bit table-data and w. are 16 bit weights
   *
   * \return unsigned 32 bit value
   */
static inline uint32_t scalar_4x4(uint16_t w1, uint16_t w2, uint16_t w3, uint16_t w4,
	uint8_t t1, uint8_t t2, uint8_t t3, uint8_t t4){
	uint32_t s;
	s = mult16_8(w1, t1);
	s += mult16_8(w2, t2);
	s += mult16_8(w3, t3);
	s += mult16_8(w4, t4);
	return s;
}

// compatibility wrapper function
// from rpm_vhr it calculates rpm_hr with formula: /10
   /*!
   * \ingroup g_math
   *
   * \brief Calculates rpm_hr with formula: /10
   *
   * compatibility wrapper function
   * from rpm_vhr it calculates rpm_hr with formula: /10
   *
   * \return unsigned 16 bit value
   */
static inline uint16_t div10(uint16_t rpm_vhr)
{	// both would work:
	return div4096(mult16_16(rpm_vhr, 410));
	// return div4096(mult16_8(rpm_vhr<<1, 205));
}

//  / 1024
   /*!
   * \ingroup g_math
   *
   * \brief Divide by 1024
   *
   * details ....
   *
   * \return unsigned 16 bit value >> 10
   */
static inline uint8_t div1024(uint16_t t)
{
	return (uint8_t)(t>>10);
}

// *41*4 / 4096
   /*!
   * \ingroup g_math
   *
   * \brief Divide by 25
   *
   * details Multiply *41*4 Divide by 4096
   *
   * \return unsigned 16 bit value
   */
static inline uint16_t div25(uint16_t t)
{
	// 163 instead of 164 is a small cheat:
	// we don't want a higher value, to prevent overflow above RX21_CONSTANT
	// see our caller
	uint8_t t163;
	uint32_t m;

	t163 = 163;
	m = mult16_8(t, t163);
	return div4096(m);
}

// no more fight with gcc to (a<<8) + b without doing + or | operations
   /*!
   * \ingroup g_math
   *
   * \brief Shift up 8 and add offset
   *
   * no more fight with gcc to (a<<8) + b without doing + or | operations
   *
   * \return unsigned 16 bit value
   */
static inline uint16_t pack8_8(uint8_t a, uint8_t b) {
	uint16_t result;
/*
%A0 to refer to the lowest byte of the first operand,
%A1 to the lowest byte of the second operand and so on...

%n0 : result
%n1 : a
%n2 : b
*/
	asm (
		"\n"
		"mov %A0, %A2 \n\t"	// b to lower  byte of result
		"mov %B0, %A1 \n\t"	// a to higher byte of result
		: "=&r" (result)
		: "r" (a), "r" (b)
	);
	return result;
}

// add the arguments, and fit into 8 bit result (properly rail)
   /*!
   * \ingroup g_math
   *
   * \brief Add 8 bit value to (16 bit value / 16)
   *
   * add the arguments, and fit into 8 bit result (properly rail)
   *
   * \return unsigned 8 bit value
   */
static inline uint8_t add8_16idiv16(uint8_t u8, int16_t i16)
{
	uint16_t u16;
	uint8_t sign;
	u16 = (uint16_t)u8;
	i16 >>= 4;	// signed (arithmetic) shift
	u16 += i16;

	sign = u16>>8;
	if(sign == 0) return (uint8_t)u16;	// fits nicely
	// oops: railing
	if(sign >= 0x80)	// negative
		return 0;
	else
		return 0xff;
}
// similar, but subtract:
   /*!
   * \ingroup g_math
   *
   * \brief Subtract (16 bit value / 4) from 8 bit value
   *
   * details...
   *
   * \return unsigned 8 bit value
   */
static inline uint8_t sub8_16idiv4(uint8_t u8, int16_t i16)
{
	uint16_t u16;
	uint8_t sign;
	u16 = (uint16_t)u8;
	i16 >>= 2;	// signed (arithmetic) shift
	u16 -= i16;

	sign = u16>>8;
	if(sign == 0) return (uint8_t)u16;	// fits nicely
	// oops: railing
	if(sign >= 0x80)	// negative
		return 0;
	else
		return 0xff;
}

// reenables interrupts for an instant so they can sneak through
   /*!
   * \ingroup g_math
   *
   * \brief Re-enable interrupts for short interval
   *
   * reenables interrupts for an instant so they can sneak through
   *
   * \return none
   */
static inline void sei_nop_cli(void)
{
	asm volatile (
		"\n"
		"sei \n\t"
		"nop \n\t"
		"cli \n\t"
		);
}
#endif
