

#ifndef __BENCHMARK_H__
#define __BENCHMARK_H__

#include <inttypes.h>

struct benchstats_t {
	uint16_t ocr0cnt;     	// B0
	uint16_t primary_trigger_cnt; // B1
	uint16_t handler2k;   	// B2
	uint16_t mainloopcnt; 	// B3
	uint16_t secondary_trigger_cnt;  // B4
	uint16_t int7cnt;     	// B5
	uint16_t kbcharlast; 	// B6
	uint16_t kbcharcnt;   	// B7
	uint16_t kbirqcnt;    	// B8
	uint16_t kbresync;    	// B9
	long unknownevent;    	// BB BA
	long eventnum;        	// BD BC
 	uint32_t highest_rpm_p;	// BF BE
	uint32_t lowest_rpm_p;	// B1 B0
	uint16_t softpwm_err; 	// B2
	uint16_t wheel_err; 	// B3
	uint16_t ign_adv_err; 	// B4
	uint16_t trig_sec_cnt;  	// B5 
	uint16_t trig_prim_cnt;  	// B6
	uint16_t ign_err;          // B7
	uint16_t engine_phase_err; // B8
	uint16_t engphase_nosync;  // B9

	uint16_t pid_overflow;  // BA
	uint8_t wbo2_overflow_Ri;	//BB
	uint8_t wbo2_overflow_nernst;	//BC
//	uint8_t wbo2_overflow_pumppid;	//BD
//	uint8_t wbo2_overflow_heaterpid;	//BE
};

#define BENCHSTATS_LENGTH sizeof(struct benchstats_t)

struct debug_t {
uint8_t *megatune_table;	// megatune can be fooled to modify other tables
uint8_t megatune_page;  // can the above be shared?
uint8_t auto_dump_freq;
uint8_t auto_dump_cnt;
uint8_t auto_dump_mask;
// uint8_t dbg_addr;
uint8_t forced_pw;
uint8_t forced_iac;
uint8_t bootl;
uint8_t ocr0_val;
uint8_t disable_flags;	// see bit defines below
uint8_t enable_flags;	// 7:auto logrepeat 6:triggerlog 4:mirror MAP=>wbo2pump 3: knock-debug 2: knock-sample-logger 1: WBO2-freeair 0: WBO2-sample-logger
// uint8_t global_tmp; // reused for several non-critical, usually non-concurrent stuff
uint16_t stim_period;	// stim 1-tooth period [8 usec]  (==pulse is low for this time [4 usec])
uint8_t stim_tooth;	// packed: 7:6 is the missing teeth to 0..3; 5:0 is toothcount 0..63 (_could_ be 3..66)
uint8_t stim_toothnow;	// what tooth we are at
uint16_t stim_nexttime;
#ifdef LED_MATRIX
uint8_t charcol;	// lower 3 bits are column
uint8_t	slow;		// counter for speed-dividor
#endif
#ifndef NOIGN
uint8_t forced_adv;
#endif
};

// number of bytes to send with mdx command
#define DBG_LEN 32

// for disable_flags
#define DIS_FLAG_LCDBIN 7
#define DIS_FLAG_AUDITRIGGERDEBUGVERBOSE 3
#define DIS_FLAG_AUDITRIGGERDEBUG 2
#define DIS_DONT_MOVE 1
#define DIS_FLAG_LCD 0

extern struct debug_t debug;

static inline uint8_t audi_trigger_debug( void )
{
	return debug.disable_flags & (1 << DIS_FLAG_AUDITRIGGERDEBUG );
}
static inline uint8_t audi_trigger_debug_verbose( void )
{
	return debug.disable_flags & (1 << DIS_FLAG_AUDITRIGGERDEBUGVERBOSE );
}

void debug_init(void);

// set this 0 to not send initially
// minimum 0, maximum 255 
#define COMM_FREQ_DEFAULT 0

#define IC3HISTOGRAM_BUCKETS 16

#ifndef EVQHISTOGRAM_BUCKETS
#define EVQHISTOGRAM_BUCKETS 1
#endif

/* each words are sent as 8 bytes plus 2 leading char "\r\n" */
#define IC3HISTOGRAM_COMMBYTES (2+IC3HISTOGRAM_BUCKETS*8)
#define EVQHISTOGRAM_COMMBYTES (2+EVQHISTOGRAM_BUCKETS*8)

extern uint16_t evq_histogram[EVQHISTOGRAM_BUCKETS];
extern uint16_t ic3_histogram[IC3HISTOGRAM_BUCKETS];
#define TMP_FOR_COMM_LEN 8
extern uint16_t tmp_for_comm[TMP_FOR_COMM_LEN];

extern struct benchstats_t benchstats;

void benchmark_init(void);
uint16_t benchmark_idle(uint16_t cycles);
uint8_t evq_latency_comm(uint8_t i);
uint8_t ic3_latency_comm(uint8_t i);
__inline__ void benchmark_histogram(uint16_t *, uint16_t, uint8_t);
uint16_t tmp_for_comm[TMP_FOR_COMM_LEN];
uint8_t tmp_comm(uint8_t i);
__inline__ void benchstats_maysend(void);

#endif 

