/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen, mik@caffrey.dk]          *
***********************************************************
*  -- History --                                          *
*    release 0.1 - 06-03-2003                             *
*            http://caffrey.dk/megasquirt                 *
*    initial Motorola version by Bowling & Grippo (2002)  *
*            http://www.bgsoflex.com/megasquirt.html      *
\*********************************************************/


/*
TODO:

There is a timeout during acceleration, this should be fixed.
Is the same timeout appearing in the B&G asm code?

*/

   /*!
   * \addtogroup g_fuel_calc Fuel Calculation Routines
   */
/*@{*/

#include "global.h"
#include "fuelcalc.h"
#include "fuelcontrol.h"
#include "helpers.h"
#include "benchmark.h"
#include "adc.h"
#include "ve.h"
#include "tables.h"
#include "eventqueue.h"
#include "injconf.h"
#include "ego.h"
#include "lcd.h"
#include "spi.h"
#include "iac.h"

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#include <string.h>
#include <avr/delay.h>
#include "arithm.c"

volatile extern struct engine_t engine;

uint8_t tpsaclk; // TPS enrichment timer clock in 0.1 second resolution

uint8_t tpsaclkcmp; // Comparison value for TPS acceleration time - from lookup table

struct corr_t corr;
struct search_table_t clt_idx;

#ifdef SOFTRPM
struct igncut_t igncut;
#endif

#include "injector.c"

/*
	progressive_delay() allows configuration of higher 
	than 25.5 msec in 8 bit config value. Used for config.cwl, cwh, injpwmt, primep
	Value 200 mean 20msec (megatune compatibility 100 usec step)
	but 201 means 20.5 msec (changing to 500 usec step)
	This allows 20 msec + 55 * 0.5 msec = 47.5 msec max pulsewidth
*/
#define PROGRESSIVE_FUNCTION_BREAK (200)
static inline uint16_t progressive_delay(uint8_t t){
	uint16_t ret;
	if(t <= PROGRESSIVE_FUNCTION_BREAK){
		return 25 * t;	// 100 usec step
	}
	
	ret = PROGRESSIVE_FUNCTION_BREAK * 25;
	t -= PROGRESSIVE_FUNCTION_BREAK;
	
	ret += t * 125; // 500 usec step
	return ret;	
}

/*
 * ONLY call from userspace
 * many squirt_t-s to fill (called from several places)
 */
void fill_all_squirters(uint16_t pw)
{
	uint8_t i, mask;
	uint16_t t;
	struct squirt_t *inj_port_p;
	// config.injpwmt is 100usec based, multiply by 25 to make it 4usec based
	inj_port_p = &inj_port[0];
	t = progressive_delay(config.injpwmt);
#ifdef INJECTOR_STAGING
	i = INJECTORS-SECONDARY_INJ_CHANNELS_NUM;
#else
	i = INJECTORS;
#endif //INJECTOR_STAGING
	do{
		i--;
		cli();
		inj_port_p->pwm_delay = t;
		inj_port_p->pwcalc = pw;
		sei();
		inj_port_p++;
		// TODO: check if gcc adds sizeof...
	} while(i);

#ifdef INJECTOR_STAGING
	// Calculating pulsewidth of secondary injectors
	uint16_t pw2;
	if(engine.tps >= config.inj_stage2_start_tps &&
		engine_kpa8() >= config.inj_stage2_start_map)
		pw2 = pw2 = mult16_16(pw, (uint16_t)(config.inj_stage2_rate))>>8;
	else
		pw2 = 0;
	i=SECONDARY_INJ_CHANNELS_NUM;
	while(i)
	{
		i--;
		cli();
		inj_port_p->pwm_delay = t;
		inj_port_p->pwcalc = pw2;
		sei();
		inj_port_p++;
		// TODO: check if gcc adds sizeof...
	}
#endif	

	// calculate injector bitmask used when firing all injectors simultaneously
	i = config.alternate & 0x0f;
	mask = injport_map(i);
	while(i) {
		i--;
		mask |= injport_map(i);
	}
	engine.injector_bitmask = mask;
}

void init_fuelcalc(void) {

  corr.ego = EGOCORR_BASE;
  corr.air = 100;
  corr.warm = 0;
  corr.baro = 100;
  corr.gammae_1k5 = GAMMAE_BASE;
  corr.ve_hp = 25600;
  corr.lambda = LAMBDACORR_STOICH;
  corr.tpsaccel = 0;
  corr.tpsfuelcut = 100;
  corr.tps_overrun = 0;
  corr.als = 0;
  corr.launch = 0;
  
#ifdef SOFTRPM
#ifdef ALS
  igncut.als = 0;
#endif
  igncut.idlecont = 0;
  igncut.revlimit = 0;
  igncut.launch = 0;
#endif // SOFTRPM
 
  fill_all_squirters(0);

  // you will have a small few msec window at about 6 hours after powerup to
  // generate fake rpm for the first trigger:
  etime.ptrig_capturedtime_prev |= 0xFF000000; // gcc: 1 byte, please
  //engine.rpm_trigger_reenable = TRIGGER_REENABLE_MIN;
}

// search the warmup_clt_range table
// TODO: only run this when engine.coolant changes
void search_clt_index(void) {
  // note, coolant is actual temperature +40 degrees F
  search_table(config.warmup_clt_range, sizeof(config.warmup_clt_range), engine.coolant, &clt_idx);
}

void leave_cranking(void)
{
	uint8_t my_status;
	cli();
	my_status = engine.status;
	if (my_status & (uint8_t) _BV(crank)) {
		my_status &= ~_BV(crank);
		my_status |= _BV(startw) | _BV(warmup);
		engine.status = my_status;
		etime.asecount = config.awc;
	}
	sei();
	
	if(config.lcd_c0 & _BV(3)){
		// start another initialization, maybe low voltage at 
		// cranking resetted the LCD
		lcd_initfrom(1);
	}
}

/***************************************************************************/
/*** warm-up and after-start enrichment                                  ***/
/*** corr.warm = 0 means *1.0 no enrichment (old 100)                    ***/
/*** corr.warm=155 means *2.55 (old max: 255)                            ***/
/*** corr.warm=255 means *3.55 (not possible before)                     ***/
/***************************************************************************/
void warmup_enrich(void) {
	uint8_t afterstart;
	uint8_t w8;	// warmup enrichment 8 bit value biased at 0
	uint16_t w;	// warmup enrichment + afterstart final calc, biased at 0

	// afterstart enrichment
	afterstart = 0;
	if (engine.status & _BV(startw)) {
		if (etime.asecount) {
			uint8_t ase_cold, ase_warm;

			ase_cold = linear_interp(0, config.awc, 0, config.awev, etime.asecount);
			ase_warm = (ase_cold * config.awev_temp_scaling) >> 8;
			afterstart = linear_interp(COLD_TEMP, WARM_TEMP, 
							ase_cold, ase_warm, engine.coolant);
		} else {
			cli();
			engine.status &= ~_BV(startw);
			sei();
		}
	}

	// no need to run through everything if the engine is warm.
	// this results in the value of wwu[last_element]is ignored.
	if (engine.coolant >= config.warmup_clt_range[sizeof(config.warmup_clt_range)-1]) {
		// outside warm-up range, clear warm-up enrichment code
		cli();
		engine.status &= ~_BV(warmup);
#ifndef GENBOARDv3
		ledstate &= ~_BV(LED_WARMUP); // wled off
#endif
		sei();

		corr.warm = afterstart;
		return;

	} else {
		cli();
#ifndef GENBOARDv3
		ledstate |= _BV(LED_WARMUP); // wled on
#endif
		engine.status |= _BV(warmup);
		sei();

		// enrichment due to coolant
		w8 = linear_interp(config.warmup_clt_range[clt_idx.lindex],
				  config.warmup_clt_range[clt_idx.uindex],
				  config.warmup_clt[clt_idx.lindex],
				  config.warmup_clt[clt_idx.uindex],
				  engine.coolant);

		// the semantics is traditionally stupid in megatune:
		// w8 is biased at 100 at this point, so we make it biased 0:
		if( w8 > 100 ){
			w8 -= 100;
		// only possible to downscale
		// the (warmup_clt[] - 100) part, with increasing rpms
		w8 = linear_interp(ve.rpm_range[0], ve.rpm_range[VE_SIZE_RPM-1], 
				w8, (w8*config.warmup_rpm_scale)>>8, engine.rpm);

		} else {
			// no enrichment (value=100) or stupid config (low-CLT leaning not applied)
			w8 = 0;
		}
	}
		
	// add the afterstart enrichment and check for overflow
	w = w8 + afterstart;

	// check overflow:
	corr.warm = (w >> 8) ? 0xFF : (uint8_t)w;
}

/*
fuelcut below a certain MAP;
return true if fuelcut required because of low MAP
*/
static inline uint8_t lowmap_fuelcut(void)
{
	uint8_t kpa8 = engine_kpa8();
	if( kpa8 < config.fuelcut_min_kpa ){
		corr.tps_overrun |= _BV(LOWMAP_FUELCUT_ACTIVE);
		return  _BV(LOWMAP_FUELCUT_ACTIVE) ;
	}

	if( kpa8 > config.fuelcut_min_kpa + LOWMAP_FUELCUT_HYSTERESIS ){
		corr.tps_overrun &= ~_BV(LOWMAP_FUELCUT_ACTIVE);
		return 0;
	}

	return corr.tps_overrun & _BV(LOWMAP_FUELCUT_ACTIVE);
}


/*
 * overrun fuelcut
 * return 1 if fuelcut is required
 */
uint8_t tps_overrun(void) {

	if( lowmap_fuelcut() ){
		return 1;
	}

	// Buck-free fuel saver!
	// When throttle is fully released the injectors will be turned off as long
	// as rpm >= fuelcut-rpm.
	// The injectors are turned on again when rpm <= fuelresume-rpm.
	// suggested values:
	//   fuelresume = 0x10 (1600 rpm)
	//   fuelcut = 0x11 (1700 rpm)

	// remember to specify config.iac_tps_thres else this won't work!
	if (engine.rpm <= config.overrun_fuelresume){
		corr.tps_overrun &= ~_BV(OVERRUN_ENABLE);
	}
	
	else if (engine.rpm >= config.overrun_fuelcut)
		corr.tps_overrun |= _BV(OVERRUN_ENABLE);


	if ( engine.tps > config.iac_tps_thres ){
		if( debug.disable_flags & _BV( DIS_DONT_MOVE ) ){
			// shut down the engine if someone steps on the TPS (antitheft)
			return 1;
		}

		// disable overrun when tps is pressed
		corr.tps_overrun &= ~_BV(OVERRUN_ACTIVE);
		// if(....) corr.ego=EGOCORR_BASE; ??
		return 0;
	} else {
		// TPS not pressed
		if(etime.asecount != 0) {
			// don't do any fuelcut during afterstart
			corr.tps_overrun &= ~_BV(OVERRUN_ACTIVE);
		        return 0;
		} else if (corr.tps_overrun & _BV(OVERRUN_ENABLE)) {
			// throttle released and rpm still way above idle
			corr.tps_overrun |= _BV(OVERRUN_ACTIVE);
			return 1;
		} else {
			if( corr.tps_overrun & ~_BV(OVERRUN_ACTIVE) ){
				// just fuel-resume 
				// drop old ego correction
				// TODO: check this
				corr.ego=EGOCORR_BASE;
				corr.tps_overrun &= ~_BV(OVERRUN_ACTIVE);
			}
			return 0;
		}
	}
}


/***************************************************************************/
/*** Throttle position / Pressure based acceleration enrichment          ***/
/***************************************************************************/
void tps_acc_enrich(void) {
  uint8_t tok_diff; // tok = tps or kpa
  uint8_t my_tok, my_last_tok;
  
  tpsaclk++;
  
  cli();
  if (config.tpsdot_kpadot_conf & _BV(tps_or_kpa)) my_tok= engine_kpa8();
  else my_tok = engine.tps;

  my_last_tok = engine.last_tok;
  sei();

    /* 
    NOTE THAT FUELCUT BASED ON 
    if(my_tok < my_last_tok)  deceleration
    WOULD BE USELESS. It was implemented until svn rev103, but it just caught the unaware !
    See fuelcut_min_kpa 
    */

  if (my_tok >= my_last_tok) {
    tok_diff = my_tok - my_last_tok;

    /* acceleration */
    if (tok_diff >= config.tpsdotrate[0]) {

      /*
	  The acceleration calculation needs to be reworked, it is unprecise.
	  It times out, and restarts computation of enrichment.
      */

      if (engine.status & _BV(tpsaen)) {
		uint8_t acc_temp_offset, acc_temp_mult, acc_enrich;
		uint16_t mtmp;
		struct search_table_t st;

		/* multiplier amount based on cold temperature 
		TODO: rip temp-calc from this.
	  	Maybe just make it able to add time when cold.
	  	But the added quantity must be scaled by corr.warm+100
	  	and req_fuel (yes!) at the end of total_pw()
		*/
		acc_temp_mult = linear_interp( COLD_TEMP, WARM_TEMP, config.acmult, 100, engine.coolant );
	
		/* lookup table amount based on tpsdot */
		search_table(config.tpsdotrate, sizeof(config.tpsdotrate), tok_diff, &st);

		acc_enrich = linear_interp(config.tpsdotrate[st.lindex],
				   config.tpsdotrate[st.uindex],
				   config.tpsaq[st.lindex],
				   config.tpsaq[st.uindex],
				   tok_diff);

		//	acc_enrich = linear_interp(st.lbound, st.ubound, config.tpsaq[st.index-1], config.tpsaq[st.index], tps_diff);

		/* calculate acc-enrichment, catch overflow */
		mtmp = mult_div100(acc_temp_mult, acc_enrich);
		if ((mtmp >> 8) != 0) // will the result fit in a byte?
		mtmp = 200;         // ... no!

		/* extra fuel due to temperature */
		acc_temp_offset = linear_interp( COLD_TEMP, WARM_TEMP, config.tpsacold, 0, engine.coolant );

		mtmp += acc_temp_offset;
		if ((mtmp >> 8) != 0) // will the result fit in a byte?
			mtmp = 0xFE;        // ... no!
	
		if (mtmp > corr.tpsaccel) 
			corr.tpsaccel = (uint8_t)mtmp; // tpsaccel is the result of this function

		if ((engine.status & _BV(tpsden)) || (tpsaclk >= tpsaclkcmp))  {
			cli();
			engine.status &= ~( _BV(tpsaen) | _BV(tpsden) );
#ifndef GENBOARDv3
			ledstate &= ~_BV(LED_ACCEL); //aled off
#endif
			sei();
			corr.tpsfuelcut = 100;
			corr.tpsaccel = 0;
		}
      } else { //start of acceleration, initialize some variables
		corr.tpsaccel = config.tpsaq[0];
		tpsaclk = 0;
		tpsaclkcmp = config.tpsasync;
		cli();
		engine.status |= _BV(tpsaen);
		engine.status &= ~_BV(tpsden);
#ifndef GENBOARDv3
		ledstate |= _BV(LED_ACCEL); //aled on
#endif
		sei();
	//	ledstate &= ~_BV(LED_WARMUP); //wled off
      }
    } else { //is acceleration done?
	  if ((engine.status & _BV(tpsden)) || (tpsaclk >= tpsaclkcmp)) {
		cli();
		engine.status &= ~( _BV(tpsaen) | _BV(tpsden) );
#ifndef GENBOARDv3
		ledstate &= ~_BV(LED_ACCEL); //aled off
#endif
		sei();
		corr.tpsfuelcut = 100;
		corr.tpsaccel = 0;
      }
    }
  }
  
  if (config.tpsdot_kpadot_conf & _BV(tps_or_kpa)) engine.last_tok= engine_kpa8();
  else engine.last_tok= engine.tps; // last tps or kpa for acc. enrichment
}

/* 
* if pw is bigger than the pwm_delay, 
* we are called to calculate the
* pwm percentage. The right way to calculate pwm percentage is
* from VBatt (similar to dwell), dispite some guys calculate
* it from rpm (cranking => low Batt), which is ugly.
* Simple interpolation according to VBatt.
* note that output range is [ v14, v14+v6 ]
* for battery voltage [14V, 6V]
* result is trimmed to 8 bits.
*/
static inline uint8_t calc_pwm_dc(uint8_t v14, uint8_t v6)
{
	uint8_t vbatt;
	uint16_t dc;

	// for anything above 14V we use the v14 value
	// (not decreased for higher than 14V voltages)
	// this is very convenient, the user normally need to set only 1 value
	dc = v14 ;
	vbatt = engine.batt;
	// TODO: use the defines, 164 is 19V!!
	// 164 (appr 14V?) and 61 (6V?) ADC readings seem to be typical high-low values
	if(vbatt<164) {	// lower Battery voltage, correction applies
		dc += ((164-vbatt) * v6 )>>7;	// at 6V: 103/128 * v6 added to v14
	}
	
	if(dc>>8) return 0xFF;
	return dc & 0xFF;
}

/*
* used from several places
*/
void static inline pwm_dc_fill(void)
{
// genboard v3 hasn't individual dutycycle setting
#ifdef GENBOARDv3
	// #ifdef FETDRIVER_INVERTING taken care at initialization
	OCR0 = calc_pwm_dc(config.injpwm, config.injpwm6);
#else
	if( inj_port[0].pwcalc > inj_port[0].pwm_delay || inj_port[1].pwcalc > inj_port[1].pwm_delay ){
		inj_port[0].pwm_dc = inj_port[1].pwm_dc = calc_pwm_dc(config.injpwm, config.injpwm6);
	}
#endif
}

/*
* Calculates cranking injector pulsewidth in 4 usec
* by doing interpolation according to coolant temperature
* between pulsewidths (progressive function of)
* config.cwl and config.cwh
* 
* Considers TPS FLOOD_CLEAR
*/
static inline uint16_t cranking_pw(void)
{
	if (engine.tps < FLOOD_CLEAR) {
		// calculate pulsewidth in 4 usec from [cwl..cwh] 
		uint16_t cwl16 = progressive_delay(config.cwl);
		uint16_t cwh16 = progressive_delay(config.cwh);
		
		return linear_interp16( COLD_TEMP, WARM_TEMP, cwl16, cwh16, engine.coolant );
		// battery correction?
	} else { 
		return 0; // flood clear
	}
}

/*
* Returns 255 if eng_bat <= minv; smaller than 255 if eng_bat is higher.
* Slope is a parameter.
* example:
* slope=0x80
* eng_bat = BATT_14V (118)
* minv = BATT_7V (61)
* result: 226 (*0.88)

* with slope=0xFF same result=198 (*0.77)
*/
static inline uint8_t vbatt_scale(uint8_t eng_bat, uint8_t slope, uint8_t minv )
{
	if (eng_bat <= minv){
		return 255;
	}

	return 255 - ((slope * (eng_bat - minv)) >> 8);
}

/* 
* Battery voltage compensation ( in 16 usec units )
* Remember, low battery voltage only delays injector opening time
* does not affect flow after inj has opened 
* and does not change closing time with constant-voltage
* (zener) flyback. With DR flyback or flyrail with
* voltage proportional to VBATT, the closing time is affected too.
* With DR flyback and lowZ the closing is not effected,
* since the pwm-duty (ideally) compensates for VBATT,
* setting a fixed current.
*
* Note that this is just the simpler part of pulsewidth extension:
* See injrampup_pw_extend()
*/
uint16_t vbatt_pw_extend(void)
{
	uint8_t batt_high, batt_low;
	uint16_t ge;

	ge = config.injopen + config.battfac; // take care of overflow
	if (ge > 0xff)
		batt_low = 0xff;
	else
		batt_low = (uint8_t)ge;

	if (config.injopen <= config.battfac)
		batt_high = 0;
	else
		batt_high = config.injopen - config.battfac;
	
// batt_low, batt_high resolution is 16 usec (max value ~4 ms)
// so config.injopen and config.battfac also (TODO: check megatun.. issues)

	return linear_interp( BATT_7V, BATT_19V, batt_low, batt_high, engine.batt );
}

/* 
* returns pulsewidth extension necessary due to rampup of injector-fuelflow.
* (in 16 usec units).

http://www.vems-group.org/index.php?page=GenBoard%2FManual%2FConfig%2FInjectorOpening

	pw: calculated fuel * req_fuel [4 usec] 
	maxext: [16 usec] is the max value used for extension (at BATT_7V) and high pw.
	For higher battery voltage return value will be downscaled linearly.

	Furthermore, for small pulsewidth, it will be downscaled according to 
	a lookup table. (that holds a very scientific squareroot-like function)
*/
uint8_t injrampup_pw_extend( uint16_t pw, uint8_t maxext){	
	uint8_t maxext_at_vbatt, index, t;

	if( maxext == 0){
		return 0;
	}
	
	// config.injocfuel [16 usec] 
	maxext_at_vbatt = (maxext 
		* vbatt_scale( engine.batt, config.injrampup_battfac, BATT_7V ))>>8;

	/*
		rampup_func_power = 1.5
		if pw is above (2/3 *maxext_at_vbatt)
		which is (maxext_at_vbatt / rampup_func_power) 
		the return value will be exactly maxext_at_vbatt

		Note that we'd use 171/256 not 42/16 if both were 16 usec
	*/
	if( pw >= ((maxext_at_vbatt * 42 )>>4) ){
		return maxext_at_vbatt;
	}

	/* if pw is lower (this is only for very small pulsewidths) 
		lookup a multiplier in a flashlookup table 
		and downscale maxext_at_vbatt accordingly.
		Note that we cannot avoid a division
		to find index (although 1/maxext_at_vbatt could be cached)

		pw canNOT be higher than 1020 (4080 usec) at this point.
	*/
#define INJOPENTABLE_SIZE 64
// or (sizeof(INJOPENTABLE[])) ?
// because of pw and rampup_at_vbatt we only need to multiply with 
#define IOTMULTIPLIER 24
// or (1.5 * INJOPENTABLE_SIZE / 4)
	index = (pw * IOTMULTIPLIER) / maxext_at_vbatt;
	// no need to check index because of above conditions
	t = PRG_RDB( &INJOPENTABLE[ index ] );
	// finally downscale:
	return ( maxext_at_vbatt * t ) >> 8;
}

/* 
* Helper function to calculate injector pulsewidth [4 usec]
* enables forced pulsewidth, 
* considers rev-limiter
* and cranking
*/

uint16_t total_pw(void)
{
	uint16_t res, active_rev_limit;
	uint8_t rpm = engine.rpm;
	
	if( config.shiftcut_conf !=0	// this should not be necessary, but somehow shiftcut_state seems to be corrupted
	&& engine.shiftcut_state & FUELCUT_MASK ) return 0;

#ifndef SOFTLAUNCH
	// config.als_rev_limit variable reused for launch control
	// emil asked that config.als_rev_limit can raise rev_limit not just decrease it
	active_rev_limit = is_als_active() ? config.als_rev_limit : config.rev_limit ;
#else
	active_rev_limit = config.rev_limit ;
#endif	

	// rough revlimiter
	if( rpm > active_rev_limit) return 0;

	// shift-cut is configured as "cut-spark" enabled
	// => we cut spark 200 RPM below fuelcut 
	if( config.shiftcut_conf & _BV( CUT_SPARK_REV )
		&& rpm > (active_rev_limit - 2)){
		engine.shiftcut_state |= _BV( IGNCUT_REVCUT );
	} else {
		engine.shiftcut_state &= ~_BV( IGNCUT_REVCUT );
	}

        // overboost protection - cut fuel if over a certain KPA
	if(config.fuelcut_max_kpa < engine_kpa8()) return 0;

	// mdp menu command can force a pulsewidth in 64 usec resolution
	if(debug.forced_pw) return debug.forced_pw << 4;
	
	if ((engine.status_ext & _BV(crank_enable)) 
		&& (engine.rpm <= config.cranking_thres)) {
		return cranking_pw();
	}  

	// overrun fuelcut
#ifdef ALS
	if (als_deactivated && tps_overrun()) return 0; // we don't want to cut fuel when ALS switch is ON!
#else // No ALS
	if (tps_overrun()) return 0;		
#endif // ALS

	// some multipliers are 100 based, stupid inheritance:
	res = mult16_8( (uint16_t)corr.warm + 100, corr.tpsfuelcut);
	// res base is now 10000. Max value is 355*100 (< 65535 fine)
	res = mult16_16(corr.baro * corr.air, res)>>16;	// res is 1526 based
	
	// other multipliers are more reasonable, based on power of 2
#if EGOCORR_BASE == 128
	res = mult16_16(res, (uint16_t)(corr.ego) + 128)>>8; // effectively (256 +- 128) / 256
#else 
#warning "EGOCORR_BASE is broken at least in comm.c and lcd_display"
#endif

#ifdef ALS
#ifdef SOFTLAUNCH
	if(corr.launch)
		res = mult16_16(res, (uint16_t)(corr.launch) + 256)>>8; // effectively (256 + 256) / 256
	else
#endif // SOFTLAUNCH
		res = mult16_16(res, (uint16_t)(corr.als) + 256)>>8; // effectively (256 + 256) / 256
#endif // ALS

	// store high precision gamma-enrichment:
	corr.gammae_1k5 = res;

	/*	up to this point all multipliers were close to 1.0 
		
		Let's apply VE, which is (256 * 100) based:
	*/
	res = div4096(mult16_16(corr.ve_hp, res));
	// res is now based on 9536.7 

	// apply the lambdacorr
	res = mult16_16(res, (uint16_t)(corr.lambda) + LAMBDACORR_OFFSET) >> 8; // effectively * 200..455 / 256

	// any remaining multiplier constant from above and below (req_fuel!!) is fixed here
	if (!(config.config13 & _BV(CONTROL_STRATEGY))){  // speed-density
		uint16_t kpa_or_tps;
		if( engine_rpm8() >= config.hybrid_rpm_m ){
			// normally, the MAP is used:
			kpa_or_tps = engine.kpa250;
		} else {
			// TODO: constant multiplier when RPM <= config.hybrid_rpm_a
			// for AlphaN, the MAP is not fully trusted
			kpa_or_tps = engine.kpa_tps_blend << 2 ;
		}
		res = mult_div596( kpa_or_tps, res );
	} else {
		res = mult16_8(res, 172) >> 8;
	}

	// TODO: res += tps_accel_enrichment * (corr.warm + 100) * const around here
	// and kill temp-related scaling from tps accel enrichment code
	// if added to res, the tpsaq[] is scaled by req_fuel (which is good!)

	// config.req_fuel is 100 usec based (little clumsy, but we stick to it)
	// we return in 4 usec resolution
	return mult16_8( res, config.req_fuel ) >> 8;
}

/***************************************************************************/
/*** calc total enrichment                                               ***/
/***************************************************************************/
void calc_total_enrichment(void) {
	uint16_t res;
	res = total_pw();
		
	// final pulsewidth calculation, wuhuw!
	if (res) {
		uint8_t batt_tmp;
		// uint16_t oc;

		// note that config.injocfuel semantics is changed
		// also note that this is called _before_ vbatt_pw_extend() applied!
		batt_tmp = injrampup_pw_extend( res, config.injocfuel );
		res += ((uint16_t)batt_tmp<<2);

		batt_tmp = vbatt_pw_extend();
		res += ((uint16_t)batt_tmp<<2) + (corr.tpsaccel*25);

	} 

	fill_all_squirters(res);
	pwm_dc_fill();
}



/***************************************************************************/
/*** calc total spark cut level                                          ***/
/***************************************************************************/
#ifdef SOFTRPM
void calc_igncut()
{
	uint8_t sparkcutlvl;

#ifdef SOFTREDLINE
	calc_revlimiter_igncut();
	sparkcutlvl = igncut.revlimit;
#else
	sparkcutlvl = 0;
#endif // SOFTREDLINE
	
	if( config.shiftcut_conf !=0	// this should not be necessary, but somehow shiftcut_state seems to be corrupted
	&& (engine.shiftcut_state & IGNCUT_MASK) != 0 )
	{
		cli();
		engine.sparkcutlvl = IGNCUTLVL_MAX; // Maximum cut level needed at shifting
		sei();
		return;
	}
	
#ifdef SOFTLAUNCH	
	calc_launch_params();
	if(igncut.launch > sparkcutlvl)
		sparkcutlvl = igncut.launch;
#endif
		
#ifdef ALS
	if(igncut.als > sparkcutlvl)
		sparkcutlvl = igncut.als;
#endif

#ifdef SOFTIDLE
	calc_idle_sparkcut();
	if(igncut.idlecont > sparkcutlvl)
		sparkcutlvl = igncut.idlecont;
#endif
		
	if(sparkcutlvl > IGNCUTLVL_MAX) sparkcutlvl=IGNCUTLVL_MAX;
		
	cli();
	engine.sparkcutlvl = sparkcutlvl;
	sei();
}
#endif //SOFTRPM

#ifdef SOFTREDLINE
void calc_revlimiter_igncut()
{
	uint16_t current_vrpm, start_rpm, current_rel_v;
	uint8_t current_rel_rpm;

	current_vrpm = engine.rpm_vhr >> 2;
	start_rpm = RV_softrpm_rpm * 25;
	
	if (current_vrpm > start_rpm){
		current_rel_v = current_vrpm - start_rpm;
		if(current_rel_v & 0xff00) current_rel_rpm = 0xff;
		else current_rel_rpm = (uint8_t)current_rel_v;
		igncut.revlimit = linear_interp(0, RV_softrpm_range, 0, IGNCUTLVL_MAX, current_rel_rpm);
	}
	else
	{
		igncut.revlimit = 0;
	}
}
#endif

#ifdef SOFTLAUNCH
void calc_launch_params(void)
{
	uint16_t current_vrpm, start_rpm, current_rel_v;
	uint8_t current_rel_rpm;

	current_vrpm = engine.rpm_vhr >> 2;
	start_rpm = RV_launch_rpm * 25;
	
	if (!launch_deactivated && current_vrpm > start_rpm){
		current_rel_v = current_vrpm - start_rpm;
		if(current_rel_v & 0xff00) current_rel_rpm = 0xff;
		else current_rel_rpm = (uint8_t)current_rel_v;
		igncut.launch = linear_interp(0, RV_softrpm_range, 0, IGNCUTLVL_MAX, current_rel_rpm);
		cli();
		engine.status_ext |= _BV(launch_active);
		sei();
#ifdef ALS
		if(!als_deactivated)
			corr.launch=RV_als_enrich;
#endif
	}
	else
	{
		igncut.launch = 0;
#ifdef ALS
		corr.launch=0;
#endif
		cli();
		engine.status_ext &= ~_BV(launch_active);
		sei();
	}
}
#endif


/*
downscale multiplier for sensor calibration 
(can be used for other sensors too - maybe use h[] table for config ?)
TODO: 10 bit ADCreading
TODO: mik, you might want to make this upscale only, or up/down
scale so less resolution is lost with newer,
more precise voltage divider standard 
(VBATT_ ... constants must be matched, and megatune issue?)
*/
static inline uint8_t batt_cal(uint8_t adcreading, uint8_t factor) {

	uint16_t t16;

	t16 = adcreading * factor;
	return (t16 >> 8);
}

// config.kpafac resolution. 2 is default, 4 (kPa) is possible. Only effects kpafac variable
#if KPAFAC_UNIT==4
#define EXPONENT_OF_512_PER_KPAFAC_UNIT 7
#else // KPAFAC_UNIT==4
#if KPAFAC_UNIT==1
#define EXPONENT_OF_512_PER_KPAFAC_UNIT 9
#else // KPAFAC_UNIT==1
	
#if (defined KPAFAC_UNIT) && (KPAFAC_UNIT!=2)
#warning "BAD KPAFAC_UNIT configured. Assuming KPAFAC_UNIT=2"
#endif

// default:
#define EXPONENT_OF_512_PER_KPAFAC_UNIT 8

#endif // KPAFAC_UNIT==1
#endif // KPAFAC_UNIT==4

// kpaoffs range is +- 32kPa
#define KPAOFS_BIAS (128)

static inline sample_t sensor11tokpa250(void) {
	sample_t k;

	if(config.kpafac != 0) {
		k = sensors11(MAP);	// max 2046..2047
// 		we certainly believe 5V railing ( eg. 255kPa, because IT HAPPENS ! )
			k = (mult16_8(k, config.kpafac)) >> EXPONENT_OF_512_PER_KPAFAC_UNIT;
			k += config.kpaoffs;
/*
			transfer function V=f(P) of mpx4250 at P=0 intercepts V in -0.04*Vs,
			sometimes config.kpaoffs=-5kPa is desirable
*/
			if( k > KPAOFS_BIAS ){
				k -= KPAOFS_BIAS;
#if EXPONENT_OF_512_PER_KPAFAC_UNIT < 8
/*
			otherwise railing not needed, ADC values limit MAP < 512 kPa
			but we don't need railing anyway, this allows control for MAP upto 1023 kPa
			if( k > 2047) k = 2047;
*/
#endif
			} else {
				k = 0;
			}
	} else {
		k = 4 * config.baro;
	}

	return k;
}

/*
 * fill engine.kpa250 from the filtered MAP sensor signal.
 * Normally
 * 0..2046 [*250 Pa]
 *=0..512 kPa (impressive boost !!)
 * but with KPAFAC_UNIT=4
 * maximum is
 * 0..4092
 *=0..1023 kPa
 */
static inline void kpafill(void){

	
	engine.kpa250 = sensor11tokpa250();
	
}	// kpafill()


/***************************************************************************/
/*** calc parameters                                                     ***/
/***************************************************************************/
void calc_parameters(void) {
  uint8_t t;

#ifdef MCP3208
  int16_t s = (int16_t) config.egt1_offs;
  uint16_t egtemp = mcp3208_data[ MCP3208_EGT1 ];
#endif

  kpafill();

  // Coolant temperature in degrees F + 40
  engine.coolant = pgm_read_byte_near(&THERMFACTOR[sensors8(CLT)]);

  // Manifold temperature in degrees F + 40
  engine.mat = pgm_read_byte_near(&MATFACTOR[sensors8(MAT)]);

  // Air Density Correction Factor
  t = pgm_read_byte_near(&AIRDENFACTOR[sensors8(MAT)]);

  // This is a hack! When idling for a long time, the intake air temperature usually rises
  // resulting in the intake air to be warm. When it gets too warm, the idling will be rough.
  // By tricking the msavr to think that the air ain't that hot, it won't lean the mixture
  // at idle rpm.
  // Suggested value of airden_ignore=62

  // We don't fiddle with the airden at cranking & warmup
  //  if (!(engine.status & (_BV(crank) | _BV(warmup))) && (t < config.airden_ignore)) {

  if (t < config.airden_ignore) {
    // normal running mode, but with low air density
    t = linear_interp(8, 14, config.airden_ignore, t, engine.rpm);
  }

  corr.air = t;

  // Interpolate throttle position
  // this makes it really smooth to install the tps
  // Should the tps be scaled 0..255 or 0..100% ?
  engine.tps = linear_interp_allow_swap(config.tps_low, config.tps_high, 0, 255, sensors8(TPS));

  // Do we want battery interpolation?
  // if the resistors in the voltagedivider are unprecise, or the vref is 
  // a little off, the batt_tmp in "calc_total_enrichment" will be wrong
  engine.batt = batt_cal(sensors8(VBATT), config.batt_cal);
  
  
  // inserted from lcd_display.c
  // no (asymetric) filter, no nothing
  // note: this is a tmp hack
#ifdef MCP3208
  if(egtemp == 0xffff){	// data not available yet
	engine.egt = 0;	// Is this correct?
  } else {
   // config.egt1_cal will be around 0x40 depending on divider resistors
	engine.egt = s + (mult16_8( egtemp , config.egt1_cal ) >> 8);
  }
  read_mcp3208( MCP3208_EGT1 );
#else
  engine.egt = 0;
#endif	// MCP3208
}


void baro_correction(void) {
	sample_t k;
	uint8_t baro_corr;
	
	// Barometric correction enabled?
	if (config.config13 & _BV(BARO_CORRECTION)) {
		/*
			baro correction is calculated the following way:
		
			barocor = 100.0 * (1.0 - 0.0047 * ((double)pressure (kpa) - 100.0));
		*/	
		
		// move back to the kPa domain instead of 250 Pa
		k =  (sensor11tokpa250() >> 2);
	
		// If the mcu resets while the engine is running, we don't believe 
		// 20 kPa being the atmospheric pressure.
		if ( (k < (config.baro - config.dbaro)) || 
		     (k > (config.baro + config.dbaro)) ) {
			
			// fill sensor BARO with a normal kpa value, but dont act 
			set_sensors8(BARO, config.baro);
			baro_corr = 100;
			cli();
			engine.status_ext |= _BV(baro_problem);
			sei();
		} else {
			uint32_t temp;
			
			// fil sensor BARO with the real kpa value
			set_sensors8(BARO, (k & 0xFF));
			if(k < 100) {
				temp = (32768 + 154*(100-k));
			}
			else {
				temp = (32768 - 154*(k - 100));
			}
			temp = temp * 100;
			temp = temp >> 15;
			baro_corr = temp;
		}
	} else {
		baro_corr = 100; // no, not enabled
	}
	
	corr.baro = baro_corr;
}

#ifdef ALS
/***************************************************************************/
/*** AntiLagSystem-induced enrichment                                    ***/
/***************************************************************************/
void als_enrich(void)
{
	uint8_t my_status, rpm, tps, time, als_sw_off, act;
	uint16_t egt;
	cli();
	my_status=engine.status;
	rpm=engine.rpm;
	tps=engine.tps;
	egt=engine.egt;
	als_sw_off=als_deactivated;
	act=(engine.status_ext & _BV(als_active));
	time=engine.als_timer;
	sei();
	
	if((my_status & _BV(crank)) ||
		als_sw_off ||
		rpm <= config.als_deact_rpm ||
		tps > RV_als_retard_maxtps ||
		(config.als_deact_time && time > config.als_deact_time) ||
		(config.als_deact_egt && egt > ((uint16_t)config.als_deact_egt << 3)))
	{
		if(act)
		{
			cli();
			engine.status_ext &= ~_BV(als_active);
			digitalout(RV_als_tks_output, DLOW);
			engine.als_timer=254;
			sei();
			corr.als=0;
#ifdef SOFTRPM
			igncut.als=0;
#endif
		}
	}
	else
	{
		if(!act)
		{
			cli();
			engine.status_ext |= _BV(als_active);
			digitalout(RV_als_tks_output, DHIGH);
			sei();
			corr.als=RV_als_enrich;
		}
#ifdef SOFTRPM
		// Calculate ignition cut level (if any)
		igncut.als=linear_interp(RV_als_cut_mintps, RV_als_cut_maxtps, RV_als_max_cut, 0, tps);
#endif
	}
}
#endif // ALS

/***************************************************************************/
/*** calc engine.rpm_vhr and engine.rpm                                  ***/
/* TODO:
 * call actuator functions from/after this.
 * Since we have previous RPM data, deltaRPM is known.
 *
 * For shiftlight, or other systems, threshold
 * against (RPM + deltaRPM_over_last_500msec) is desired
 * People usually say they want lower shiftlight threshold
 * in gear1 than in gear5, but what they really need is this.
 *
 * simple solution:
 * threshold > (newRPM * const1 - prevRPM * const2)
 *
 * advanced solution:
 * timespent since previous calc_rpm() 
 * which is actually == rpm_period
 * _might_ be used in the calc,
 * const1 and const2 could depend on it.
 * since time of previous RPM depends mostly on RPM,
 * for shiftlight and revlimit the simple solution
 * can work very well.
 *
 * example:
 * t=500msec (reaction time. Maybe 400msec? including shifting? unlikely)
 * timespent=5msec (6000 RPM=10msec/crankrot, 4 cyl)
 * threshold comparison against: newRPM + (newRPM-prevRPM) * t/timespent
 * since t/timespent = 100 = rpm_vhr/60
 * const1=101
 * const2=100
 *
 * 1) in any case, take care about overflow
 * 2) if(prevRPM >= newRPM) ... compare threshold against newRPM without any wizardry.
*/
/***************************************************************************/
void calc_rpm(void) {
	uint16_t rpm, rpmk, rpm_p;
	uint32_t rpm_p32, rpmk32;

	// make sure the bytes are fetched atomically
	cli();
	rpm_p32 = engine.rpm_period;
	sei();

#ifdef BENCHMARK
	if (rpm_p32 < benchstats.lowest_rpm_p)
		benchstats.lowest_rpm_p = rpm_p32;
	if (rpm_p32 > benchstats.highest_rpm_p)
		benchstats.highest_rpm_p = rpm_p32;
#endif


	if((uint16_t)(rpm_p32 >> 16)){	// very small RPM
		cli();
		engine.rpm_vhr = 100;
		sei();
		engine.rpm = 1;
		return;
	}
    
	// rpm_p is the number of 4 usec's in a tach-event
	rpm_p = rpm_p32 & 0xFFFF;

	// rpm_p below config.crank_minper [16 usec resolution] is considered fake
	// this also prevents division by zero below
	if(rpm_p < ((uint16_t)(config.crank_minper) << 2)){	// abnormally small period
		if (rpm_p) {  // silly trigger 
			engine.rpm = 255;			// 25500 RPM
			cli();
			engine.rpm_vhr = 25500;	// fake value
			sei();
		} else {	// seems there were no valid rpm_p values yet
		// Note that the writeback of rpm_vhr is atomic. 
		// Since we do not call this from interrupt, only from userspace:
		// cli()-sei() is fine (no SREG saving needed)
			engine.rpm = 0;
			cli();
			engine.rpm_vhr = 0;
			sei();
		}
		return;
	}


	rpmk = (config.rpmk[0] << 8) | config.rpmk[1];  // I think I'll set it 3000 (0x0BB8) for my 4 cyl
	
	// MIK: The following multiplications must be type casted to 32bit, else the
	// upper two bytes of the result will be truncated.

	// 25 since rpm_p is in 4 usec not 100. 100 because we want real rpm, not [100 rpm] resolution
	rpmk32 = mult16_16(rpmk, 25*100);
	
	// rpm_hr was the number of rotations in 6 sec (real RPM/10)
	// no division by zero, because of checks above
	rpmk = rpmk32 / rpm_p;	// real rpm
	cli();
	engine.rpm_vhr = rpmk; 
	sei();
	// division by 100: * 41 / 4096
	// think about rounding. What do we want? Maybe rpmk += 50 before truncating?
	rpm = div4096(mult16_8(rpmk, 41));	// faster and more safe than trying 8*8 mult

	if ((rpm >> 8) != 0)
		engine.rpm = 255; // 25.500 rpm should be enough for anybody (Bill Gates '81)
	else
		engine.rpm = rpm & 0xFF;

}


/***************************************************************************/
/*** cranking                                                            ***/
/***************************************************************************/
void cranking(void) {
  uint8_t my_status;

  cli();
  my_status = engine.status;
  my_status |= _BV(crank);
  my_status &= ~( _BV(startw) | _BV(warmup) );
  engine.status = my_status;
  sei();
}


/**************************************************************************
* primepulse
* Call under cli()
***************************************************************************/
void primepulse(void) {
	uint16_t primep_cold, primep_hot, pw;
	uint8_t primep;

	primep = config.primep;
	if (primep == 0){ 
		// BEWARE: fill_all_squirters() not called (injpwmt OK?)
		// TODO: is that OK?
		return;
	}
	
    // don't prime if the engine is flooded
	if (engine.tps >= FLOOD_CLEAR)
		return;
  
	primep_cold = progressive_delay(primep);
	primep_hot = mult16_8(primep_cold, config.primep_temp_scaling) >> 8;
	
	pw = linear_interp16( COLD_TEMP, WARM_TEMP, primep_cold, primep_hot, engine.coolant );
      
	// do we want battery correction?

	pwm_dc_fill();

	if (pw == 0) { // if the pulsewidth turned out to be zero
		return;
	}
	
	fill_all_squirters(pw);

	/* the primepulse is handled similarly as any injection event-sequence 
	* we are in the initial cli() condition, so no need for special care
	*/

	readTMR();

#ifdef GENBOARDv3
	injector_start_all();
#else
	injector_start(&inj_port[0], startinj1, INJAGOPWM, INJASTOP);
	injector_start(&inj_port[1], startinj2, INJBGOPWM, INJBSTOP);
#endif
	alarm2core();
}	// primepulse()

/*@}*/

