/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen, mik@caffrey.dk]          *
***********************************************************
*  -- History --                                          *
*    release 0.1 - 06-03-2003                             *
*            http://caffrey.dk/megasquirt                 *
*    initial Motorola version by Bowling & Grippo (2002)  *
*            http://www.bgsoflex.com/megasquirt.html      *
\*********************************************************/


#ifndef FUELCALC_H
#define FUELCALC_H

#include <inttypes.h>


struct corr_t {
  uint8_t ego;       // Oxygen Sensor Correction
  uint8_t air;       // Air Density Correction lookup
  uint8_t warm;      // Total Warmup Correction
  uint8_t baro;      // Barometer Lookup Correction
  uint16_t gammae_1k5;    // Total Gamma Enrichments, 1526 based
  uint16_t ve_hp;     // Current VE value & fraction from lookup table
  uint8_t lambda;    // lambda (afr)
  uint8_t tpsaccel;  // Acceleration enrichment
  uint8_t tpsfuelcut; // fuelcut
  uint8_t tps_overrun; // tps overrun fuelcut flags
  uint8_t als; // antilagsystem enrichment
  uint8_t launch; // antilagsystem enrichment
};

#ifdef SOFTRPM
struct igncut_t {
#ifdef ALS
  uint8_t als; // antilagsystem igncut
#endif
  uint8_t idlecont; // idle control igncut
  uint8_t revlimit; // rev limiter igncut
  uint8_t launch; // launch control rev limiter igncut
};
#endif // SOFTRPM


#define GAMMAE_BASE (1526)

extern uint8_t tpsaclk;
extern struct search_table_t clt_idx;
extern struct corr_t corr;
extern struct igncut_t igncut;

void leave_cranking(void);
void fill_all_squirters(uint16_t pw);
void init_fuelcalc(void);
void search_clt_index(void);
void calc_parameters(void);
void baro_correction(void);
void warmup_enrich(void);
void tps_acc_enrich(void);
void o2(void);
void calc_total_enrichment(void);
#ifdef SOFTRPM
void calc_igncut(void);
#endif
uint8_t tps_overrun(void);
#ifdef ALS
#define als_deactivated (ALS_SWITCH_IN & _BV(ALS_SWITCH_PIN))
void als_enrich(void);
#endif // ALS

#ifdef SOFTREDLINE
void calc_revlimiter_igncut(void);
#endif
#ifdef SOFTLAUNCH
#define launch_deactivated (LAUNCH_SWITCH_IN & _BV(LAUNCH_SWITCH_PIN))
void calc_launch_params(void);
#endif

void calc_rpm(void);
void cranking(void);
void primepulse(void);


// tps_overrun fuelcut bit flags
#define OVERRUN_ENABLE 0
#define OVERRUN_ACTIVE 1
#define LOWMAP_FUELCUT_ACTIVE 2

#ifndef LOWMAP_FUELCUT_HYSTERESIS
#define LOWMAP_FUELCUT_HYSTERESIS 2
#endif

#endif
