/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Marcell Gal cell at x-dsl.hu]                *
***********************************************************
* used for injection scheduling (included, not linked)    *
\*********************************************************/
  /*!
   * \addtogroup g_fuel_control Fuel Injector Control Routines
   */
/*@{*/

#ifndef __INJECTOR_C__
#define __INJECTOR_C__
#include "global.h"
#include "ve.h"
#include "eventqueue.h"
#include "lbus.c"
#include "dispatcher.h"
#include "heap.h"

/* 
* starts injection and checks if gopwm or injstop needs to be scheduled
* Caller must call readTMR() before this function, and be in cli()
* Call alarm2core() afterwards
*/

#ifdef GENBOARDv3

/*	TODO: make these configurable.
	Note that you can easily do alternate
	for now, respect mik's soldering:
	
	PORTA bits 0,5,6,7 channels thus A,F,G,H will look like:
	static uint8_t injport_hw_map[8] = {0x80, 0x40, 0x20, 1, 0,0,0,0};
	and config.alternate=03

	For first tests, 2 channels are fired simultaneously:
	config.alternate=01
	PORTA bits 0,5 first, than 6,7 
*/
// static uint8_t injport_hw_map[8] = {0xC0, 0x21, 0,0, 0,0,0,0};

// injport_map is in hwmap line zero:
#define injport_map(x) (hwmap.table[x])
// hwmap line1 are individual values
//#define CRANKINJ_MASK (hwmap.table[HWMAP_SIZE])
/* TODO: these will not do what you want 
	with the inverting MC34151D FETdrivers.
	Maybe 
#define INJPOLARITY_BITS (hwmap.table[HWMAP_SIZE+1])
	The ioinit() can be done by setting relevant pins as input,
	and the HW pullup/down will determine
*/
static inline void startinj(uint8_t upper_v)
{	
	uint8_t mask;
	mask = injport_map(upper_v);
#ifdef FETDRIVER_INVERTING
	INJ_PORT &= ~mask;
#else
	INJ_PORT |= mask;
#endif
	INJ_DDR  |= mask;
}

static inline void stopinj(uint8_t upper_v)
{
	uint8_t mask;
	mask = injport_map(upper_v);
#ifdef FETDRIVER_INVERTING
	INJ_PORT |= mask;
#else
	INJ_PORT &= ~mask;
#endif
	INJ_DDR  |= mask;
}

static inline void pwminj(uint8_t upper_v)
{
	uint8_t mask;
	mask = injport_map(upper_v);
#ifdef FETDRIVER_INVERTING
	// yes, same operation! we don't want pullups.
	INJ_PORT &= ~mask;
#else
	INJ_PORT &= ~mask;
#endif
	INJ_DDR  &= ~mask;
}


static inline void injector_start(uint8_t i)
{
	uint16_t pwcalc;
	struct squirt_t *inj_port_i_p;
	inj_port_i_p = &inj_port[i];
	pwcalc = inj_port_i_p->pwcalc;
		
	// note, that we do not change inj_port.stopat if we do not schedule
	// (pwcalc==0). That might disturb the turnoff of injector.
	if (pwcalc){
		uint16_t newtime, pwd, stopat;
		val_t v;
		
		pwd = inj_port_i_p->pwm_delay;
		newtime = etime.now;

		v = v_shiftup(i);	// start with upper bits
		startinj(i);

		stopat = newtime + pwcalc;
		inj_port_i_p->stopat = stopat;

		if(pwd < pwcalc){ // PWM active
			newtime += pwd;
			v += INJ_PWM;
		} else { // PWM not active				
			newtime = stopat;
			v += INJ_STOP;
		}
	
		event_insert(newtime, v);
	}
}	// injector_start()



static inline void startinj_all(void)
{
	uint8_t mask = engine.injector_bitmask;
#ifdef FETDRIVER_INVERTING
	INJ_PORT &= ~mask;
#else
	INJ_PORT |= mask;
#endif
	INJ_DDR  |= mask;
}

static inline void stopinj_all(void)
{
	uint8_t mask = engine.injector_bitmask;
#ifdef FETDRIVER_INVERTING
	INJ_PORT |= mask;
#else
	INJ_PORT &= ~mask;
#endif
	INJ_DDR  |= mask;
}

static inline void pwminj_all(void)
{
	uint8_t mask = engine.injector_bitmask;
#ifdef FETDRIVER_INVERTING
	// yes, same operation! we don't want pullups.
	INJ_PORT &= ~mask;
#else
	INJ_PORT &= ~mask;
#endif
	INJ_DDR  &= ~mask;
}

static inline void injector_start_all(void)
{	
	uint16_t pwcalc;
	pwcalc = inj_port[0].pwcalc;
	if (pwcalc){
		uint16_t newtime, pwd, stopat;
		val_t v;
		
		pwd = inj_port[0].pwm_delay;
		newtime = etime.now;

		startinj_all();

		stopat = newtime + pwcalc;
		inj_port[0].stopat = stopat;

		if(pwd < pwcalc){ // PWM active
			newtime += pwd;
			v = INJ_ALL + v_shiftup(INJ_ALL_PWM);
		} else { // PWM not active				
			newtime = stopat;
			v = INJ_ALL + v_shiftup(INJ_ALL_STOP);
		}
		event_insert(newtime, v);
	}
}


// dispatcher must destroy heap_normal.heaproot
static inline void injector_start_destroy(uint8_t i)
{
	uint16_t pwcalc;
	struct squirt_t *inj_port_i_p;
	inj_port_i_p = &inj_port[i];
	pwcalc = inj_port_i_p->pwcalc;
		
	// note, that we do not change inj_port.stopat if we do not schedule
	// (pwcalc==0). That might disturb the turnoff of injector.
	if (pwcalc){
		uint16_t newtime, pwd, stopat;
		val_t v;
		
		pwd = inj_port_i_p->pwm_delay;
		newtime = etime.now;

		v = v_shiftup(i);	// start with upper bits
		startinj(i);

		stopat = newtime + pwcalc;
		inj_port_i_p->stopat = stopat;

		if(pwd < pwcalc){ // PWM active
			newtime += pwd;
			v += INJ_PWM;
		} else { // PWM not active				
			newtime = stopat;
			v += INJ_STOP;
		}

		event_insert_destroy(newtime, v);

	} else {
		heap_del1st(heap_normal);
	}// else maybe stopinj()? INJ_STOP scheduled for sure, 
	// not urgent, actually better this way (no lean fire,
	// if properly timed...
}	// injector_start_destroy()

#else

static inline void injector_start(struct squirt_t *inj_port, 
	void (*startinj)(void), val_t INJGOPWM, val_t INJSTOP)
{
	uint16_t now = etime.now;
	// note, that we do not change inj_port.stopat if we do not schedule
	// (pwcalc==0). That might disturb the turnoff of injector.
	if (inj_port->pwcalc){
		startinj();
		inj_port->stopat = now + inj_port->pwcalc;
		if(inj_port->pwm_delay < inj_port->pwcalc){ // PWM active
			event_insert(now + inj_port->pwm_delay, INJGOPWM );
		} else { // PWM not active				
			event_insert(now + inj_port->pwcalc, INJSTOP );
		}
	} 
}

#endif	// GENBOARDv3

/* 
  BEWARE: call between cli() / sei()

  state is DHIGH (0x80) or DLOW (0) 
  (do we want state to be packed into selector by caller?)
  selector's bit7:4 is the same semantics as l259_packed()'s valbit
  selector's bit3:0 is a facility (like S259, I259...)
  
  selector MUST match make_conf.pl varvaldictionary() and wiki descriptions.
  
  TODO (not needed too much): assign some bit-combinations for dual-output 
  (like STEPPER_B, STEPPER_D) with sei_nop_cli(); in between
  
  Note that INJFET_n and INJFETi_n references h[0,n],
  so several FETs can be switched on.
*/
static inline void digitalout(uint8_t selector, uint8_t state)
{
	uint8_t *cs_port_addr;
	uint8_t valbit;
	// facility is in the lower 4 bits, currently 3 bits are used:
	uint8_t facility = selector & 0xf;
	
	valbit = selector & 0xf0;
	valbit ^= state;
	
	// no cli();  !!! although this would be later than what the 
	// caller can do, cli() here doesn't worth the extra caution.
#ifdef GENBOARDv3
/*
	%varvaldictionary =	(
	INJFET => 0,
	SPECFET => 1,
	I259 => 2,	
	S259 => 6,
	P259 => 7,
	DISABLED => 15
	);
*/
	switch(facility){			
		case 0:
			// the injectors are addressed in the 3:0 bits, unlike the l259_packed
			selector = (valbit & 0x70) >> 4;
			if(valbit & DHIGH)
				startinj(selector);
			else
				stopinj(selector);
		
			return;
		// SPECFET: warning: if PORTB5 / PORTB6 is controlled by OCR, this won't work:
		case 1: 
				if(valbit & 0x80){
					PORTB |= _BV( (valbit & 0x70) >> 4 );
				} else {
					PORTB &= ~ _BV( (valbit & 0x70) >> 4 );		
				}		
			return;
		case 2: cs_port_addr = (uint8_t *)&CS_IGN259_PORT; break;
		case 6: cs_port_addr = (uint8_t *)&CS_S259_PORT; break;
		case 7: cs_port_addr = (uint8_t *)&CS_P259_PORT; break;
		default: 
		// eg. the FF value will do nothing:
			return;
	}

	l259_packed(valbit, cs_port_addr, facility);
#else
	// TODO: what to do on v2 ?? Move all actuation here ?
	if(facility != 0xf){
		if(state)
			PORTB |= _BV(IDLEVALVE);
		else
			PORTB &= ~_BV(IDLEVALVE);
	}
#endif
	// no sei();  !!!
}

#endif	// __INJECTOR_C__

/*@}*/
