/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Marcell Gal cell at x-dsl.hu]                *
***********************************************************
* io initialization is split out, because also used from 
* bootloader 
* not linked, but inlined (linking across bootloader and
* megasquirt would be dangerous anyway :-)
\*********************************************************/

#ifndef __IOINIT_C__
#define __IOINIT_C__

/* 
* initializes IO ports
* used for GenBoard
*/

// #include "edis.c"
/*!
 * \ingroup g_init
 * \brief I/O Port Initialization for power-up.
 *
 * \return none
 */

static inline void ioinit(void)
{
#ifdef GENBOARDv3
#if (defined __MEGASQUIRT_C__) || (! defined HARDWARE_PULLUP_OR_DOWN)
// when called from megasquirt.c we always set output
// note: firmware can be uploaded easier then bootloader!

// For bootloader, HARDWARE_PULLUP_OR_DOWN means that
// the SIL5 resistor networks installed and R154 or R153
// is used depending on FETDRIVER (inverting or not)
// than we do nothing: leave the port in high impedance (input)
#ifdef FETDRIVER_INVERTING
	PORTA = 0xFF;
#else
	PORTA = 0x00; // no pullup for injectors
#endif // FETDRIVER_INVERTING
	DDRA = 0xFF;
#endif // HARDWARE_PULLUP_OR_DOWN

	// SS high
	PORTB = 0x01;
	DDRB = 0xF7;

	// CS_KNOCK high
	PORTC = 0x04; 
	DDRC = 0xFF;	// a bad LCD_DATADDR30 can screw this !

	// CS_S259, CS_P259, CS_CAN are high
#ifdef FORCE_TRIGGER_IC1_PULLUP
	PORTD = 0xF0;   // pullup on IC1 (0x10) on
#else
	PORTD = 0xE0;	// pullup on IC1 (0x10) off
#endif
	DDRD = 0xEB;

/*
#ifdef AUDITRIGGER
    // HALL signal is NOT connected to PIND0 auditrigger setup
    //   - turn pullup PD0 (0x01) on
    //   - direction = input
    //   - Configure interrupt for falling edge
    PORTD |= 0x01;
    DDRD &= ~0x01;
    EICRA |= _BV(ISC01);  // page 86.
    EICRA &= ~_BV(ISC00);  // Set interrupt on falling edge
    EIFR = _BV(INTF0);  // clear interrupt flag
    EIMSK |= _BV(INT0);  // page 88. enable interrupt
#endif
*/

#ifdef FORCE_TRIGGER_IC3_PULLUP
	PORTE = 0x80; // secondary trigger PE7=IC3 pullup on
#else
	PORTE = 0x00; // secondary trigger PE7=IC3 pullup off
#endif	// TRIGGER_PULLUP_OFF

	DDRE = 0x3A;

	PORTF = 0x00; // adc channels
	DDRF = 0x00;

	// outputs CS_MCP(4), CS_FLASH(3), CS_IGN259(2) are high, HEAT2(1), HEAT1(0) are low:
	PORTG = 0x1C;
	DDRG = 0x1F;
	
#else	// genboard v2.2
	PORTA = 0x0F; // misc LEDs
	DDRA = 0xF8;	// see LCD_DATADDR30
#ifdef FETDRIVER_NONINVERTING
	PORTB = 0x03;	// pwm active high
#else
	PORTB = 0x63;	// pwm active low
#endif
	DDRB = 0xFC;

	PORTC = 0xE0; // stepmotor
	DDRC = 0x1F;

	PORTD = 0xEF;
	DDRD = 0x10;
#ifdef TRIGGER_PULLUP_OFF
	PORTE = 0x2F;   // this disables the pullup on PE7=IC3, should not be necessary, see http://www.squirrelpf.com/msavr/index.php?page=DocsAndFAQ%2FEDISFAQ
#else
	PORTE = 0xAF;   // we want pullup at the TACH in to prevent noise if nothing connected
#endif	// TRIGGER_PULLUP_OFF

	// wbo2 heater low
	DDRE = 0x18;	// output 0x18 in injconf, 0x20 in ignconf

	PORTF = 0x00; // adc channels
	DDRF = 0x00;

	PORTG = 0x05;
	DDRG = 0x1B;
//	ign_init();	// TODO: fix for v2 and v3
#endif

#ifdef ALS
	ALS_SWITCH_DDR &= ~_BV(ALS_SWITCH_PIN); // define als switch as input pin // TODO: some 'check against morons'?
	ALS_SWITCH_PORT |= _BV(ALS_SWITCH_PIN); // Pullup activated for ALS switch
#endif

#ifdef SOFTLAUNCH
	LAUNCH_SWITCH_DDR &= ~_BV(LAUNCH_SWITCH_PIN); // define LAUNCH switch as input pin // TODO: some 'check against morons'?
	LAUNCH_SWITCH_PORT |= _BV(LAUNCH_SWITCH_PIN); // Pullup activated for LAUNCH switch
#endif

}

#endif
