// menu definitions and signatures
#ifndef __MENU_H__
#define __MENU_H__

#include "inttypes.h"

// bits of pos_mode:
#define POS_AUTO	7

struct menu_t {
	uint8_t state;	// where are we?
	uint8_t table;	// selected table 'n' ignition, 'j' injection
	uint8_t dimension;	// cached dimension for the above table lastline_idx <<4 + lastcol_idx (0xf7 = 16x8 table)
	uint8_t *tab;		// cached table-pointer (direct function of table)
	uint8_t pos_mode;	// 0: normal mode; otherwise (>0x80) pos retreived from realtime engine parameters
	uint8_t posh;	// posh:posl explicit addressed position (OFFSET calculated from these) 
	uint8_t posl;	// inside table (table-dependent, usually line=posh and column=posl )
	uint8_t num;	// number that is halfway input
	uint8_t output;	// selects output device 0:none, 1:LCD, 2:serial vt100
	uint8_t options; // misc options, 0:display current 'bin'
};

#define SHOW_BINS 0

extern struct menu_t menu_rs232;
#ifdef PS2KEYBOARD
extern struct menu_t menu_ps2;
#endif
/* currently a flat menu is implemented that is controlled by 1 byte.
* Performance is OK: gcc -O2 can optimize a huge switch-case.
* in the future when state numbers grow, a 2 deep could be better
*/

// enum for menu state;
// for RS232 old-megatune mode is MENU_ZERO
// after you type "Man" in RS232 it goes to mmm (main menu)
// for ps2 keyboard we start in mmm
// IMPORTANT: leave MENU_can* at the very end!
enum { MENU_ZERO, MENU_ZERO_M, MENU_ZERO_Ma, MENU_mmm, MENU_mt, MENU_mtt, MENU_mtt_g,
MENU_mtt_gx, MENU_b, MENU_by, MENU_mtt_x, MENU_mtt_xx, MENU_mx, MENU_ml, MENU_md,
MENU_num, MENU_num_x, MENU_mc, MENU_mcS, MENU_mcL, MENU_mi, MENU_mv, MENU_ms, MENU_sipr, 
MENU_sipr_inc, MENU_sipr_M, MENU_sipr_Ma, MENU_can, MENU_can_inc, MENU_can_inc_escape };

/* ------ return values for menu_char() for future usage, when we want to 
distignuish between successful processing and unhandled characters
*/
#define MENU_RET_OK
#define MENU_RET_UNHANDLED

void menu_init(struct menu_t *, uint8_t);

/* call menu_char() whenever a character is ready to be processed, from 
userspace buffer handling (not good to process in interrupt anyway).
For rs232 it calls comm() if in old megatun.. mode when appropriate.
*/
void menu_char(struct menu_t *, uint8_t);


enum { 
MNUM_DIGITALOUT=1
,MNUM_IGNDWELL	
,MNUM_SRLDEBUG
,MNUM_MCB
,MNUM_MSP
,MNUM_MLP
};

#endif

