

#include <avr/crc16.h>

// needed for NULL
#include <stdlib.h>
#include <inttypes.h>

#include "global.h"
#include "sipr_protocol.h"
#include "sipr.h"
#include "comm.h"
#include "menu.h"
#include "slcan.h"

uint8_t sipr_incoming_recCRCl;

uint8_t srl_sipr_packet( uint8_t i)
{
  uint8_t c;  // char we will send
  struct cmd_queue_t *cmdqh;
  cmdqh = &cmd_queue[cmd_queue_head];
  uint8_t len=cmdqh->len;  // length of this packet

  switch (i) {
    case 0: // header byte return directly to avoid applying CRC
      cmdqh->privw=0xffff;  // the CRC will be stored here
      return SIPR_HEADER;
    case 1: // sequence number
      c=cmdqh->seq;
      break;
    case 2: // packet type always found at func(0)
      if (cmdqh->privatefunc!=NULL)
        c=cmdqh->privatefunc(0); // callout for the packet content
      else
        c='X'; // something BAD happened!
      break;
    case 3: // packet data length calculated from the length subtracting the headers and CRC
      c=len-6;
      break;
    default: // process any data and the CRC bytes
      if (i==len-2) // CRC byte 1
        return LO( cmdqh->privw );
      else 
        if (i==len-1) // CRC byte 2
          return HI( cmdqh->privw );
        else
          if (cmdqh->privatefunc!=NULL)
            c=cmdqh->privatefunc(i-3); // callout for the packet content
          else 
            c='X'; // something BAD happened!
  }
  // update the CRC
  cmdqh->privw = _crc16_update(cmdqh->privw, c);
  return c;
}

/**********************************************
* function for adding SIPR send functions. Different from normal enqueue commands...
* this hard codes the call to srl_sipr_packet and sets the function as the packet
* specific content generator. The last number passed is the packet sequence number
*/
uint8_t pushfuncsipr(uint8_t (*f)(uint8_t i), uint8_t len, uint8_t seq) 
{
  uint8_t next_tail;

  next_tail = (cmd_queue_tail + 1) % CMD_QUEUE_SIZE;
  if (next_tail == cmd_queue_head)
    return -1; // whoops, queue is full

	// depending on menu state, "packet formatter" is srl_slcan_packet or srl_sipr_packet
  	if( menu_rs232.state >= MENU_can ){
  		cmd_queue[cmd_queue_tail].func = &srl_slcan_packet;
  		cmd_queue[cmd_queue_tail].len = len + 3;	// slcan is more verbose
	} else {
  		cmd_queue[cmd_queue_tail].func = &srl_sipr_packet;
		cmd_queue[cmd_queue_tail].len = len;
	}

  cmd_queue[cmd_queue_tail].count = 0;
  
  // privatefunc is the same for slcan/sipr
  cmd_queue[cmd_queue_tail].privatefunc = f;
  cmd_queue[cmd_queue_tail].seq = seq;
  cmd_queue_tail = next_tail;

//  UCSR0B |= _BV(UDRIE0); // enable transmit interrupt
  return 0;
}


/*****************************
* send
void sipr_crc_error( void )
{
	pushfuncsipr(&sipr_packet_NAK, 6, sipr_incoming.seq);
        // we are no longer waiting for a packet...
        sipr_inc_timeout=255;  
}
*/

/* processes incoming chars */
void sipr_process_inc(uint8_t c) {
  /* reset the timeout */
  sipr_inc_timeout=2;

  sipr_incoming.idx++;
  switch (sipr_incoming.idx) {
    case 1:
      sipr_incoming.CRC=0xffff;  // don't forget to reset the CRC value because this is a new packet
      sipr_incoming.seq=c;
      break;
    case 2:
      sipr_incoming.type=c;
      break;
    case 3:
      sipr_incoming.size=c;
      break;
    default:
      if (sipr_incoming.idx==sipr_incoming.size+4) {  // CRC byte 1
        sipr_incoming_recCRCl = c;
        return;
      }
      if (sipr_incoming.idx==sipr_incoming.size+5) { // CRC byte 2 and check
      	uint16_t crc16 = sipr_incoming.CRC;
        // CRC mismatch (either the lower or the upper byte) => send NAK
        if ((sipr_incoming_recCRCl != LO(crc16))
        	|| (c != HI(crc16)) ) {
          pushfuncsipr(&sipr_packet_NAK, 6, sipr_incoming.seq);
        }
        else { // valid packet
          sipr_process_packet();
        }
        // we are no longer waiting for a packet...
        sipr_inc_timeout=255;  
        return;
      }
      else {
        // accept and store only MAX_INC_SIZE chars
        if (sipr_incoming.idx-4<MAX_INC_SIZE)
          sipr_incoming.data[sipr_incoming.idx-4]=c;
      } 
         
  }  // switch

  /* update the CRC */
  sipr_incoming.CRC=_crc16_update(sipr_incoming.CRC, c);
}
