/* Copyright 2005 (C) by Marcell Gal
 * All rights reserved.  
 */

#include <avr/crc16.h>
#include <stdlib.h>

#include "global.h"
#include "comm.h"
#include "slcan.h"
#include "sipr.h"

/*
* Sends out a CAN packet and calculates the CRC16
* on the ORIGINAL PACKET (production data, not wire-data)
* _BEFORE_ byte stuffing (of course, since the CRC16
* might need byte stuffing itself!)
* http://wiki.x-dsl.hu/cgi-bin/w/telemetry.cgi?UartPacketComm
*/
uint8_t srl_slcan_packet( uint8_t i){
  struct cmd_queue_t *cmdqh;
  cmdqh = &cmd_queue[cmd_queue_head];
  uint8_t c;  // char we will send

  uint8_t len=cmdqh->len;  // length of this packet

  switch (i) {
    case 0:
      cmdqh->privw = CRC_INITIAL;
        // return directly to avoid applying CRC
      return HDLC_END;
    case 1:
    	// TYPE and length
   		// bit7=0 means this is the last frame TODO: 1 when applicable
    	// bit6:5 type=CAN (giga8, IPV4 and IPV6 reserved, not supported for now)
    	// bit4:0 length of the 
    	// packet data length calculated from the length subtracting the headers and CRC
      	c = len - 8;
      	break;
    case 2:
    
    	// CAN-ID 4-byte word as LSB first (unlike in CAN)
    	c = node_address_low();
      	break;
    case 3:
       	c = node_address_high();
      	break;
    case 4:
		// packet subtype (SIPR type) always found at func(0)
     	// currently in the range 0..9
     	// nested in the CAN-ID
      	if (cmdqh->privatefunc != NULL)
        	c=cmdqh->privatefunc(0); // callout for the packet content
      	else
        	c='X'; // something BAD happened!
      	break;
    case 5:
    	// constant for SIPR;
    	// relatively low priority: only 30..31 are lower priority
      	c = CANID_SIPR;
      	break;
    case 6:
    	// Start of CAN data
      	// SIPR sequence number wired in as first byte
      	// for SIPR functional compatibility
      c = cmdqh->seq;
      break;
    default: // process any data and the CRC bytes
      if (i == (len-2)) // CRC byte 1
        return LO( cmdqh->privw );
      else 
        if (i == (len-1)) // CRC byte 2
          return HI( cmdqh->privw );
        else
          if (cmdqh->privatefunc != NULL)
            c=cmdqh->privatefunc(i-6); // callout for the packet content
          else 
            c = 'X'; // something BAD happened!
  }
  // update the CRC
  cmdqh->privw = _crc_ccitt_update(cmdqh->privw, c);
  return c;
}

/*
 * receive CAN packet from wire
 * first: unescape ...
 */
void slcan_process_inc( uint8_t c )
{
	uint8_t idx = sipr_incoming.idx;
	idx ++;
	sipr_incoming.idx = idx;

	switch ( idx ) {
		// HDLC_END only results in menu state transition, never seen
    case 1:
      	sipr_incoming.CRC = CRC_INITIAL;
      	sipr_incoming.size = c & 0x1F;	// TODO: support only CAN with length 0..8 for now ?
      	break;
    case 2:	// address low
    	sipr_incoming.type = c;	// temporary location for address-low byte
      	break;
    case 3: // address high;
      	if(
     	 ( (c == node_address_high()) && (sipr_incoming.type == node_address_low()) ) // our address
      		|| (( c == 0) && (sipr_incoming.type == 0 ))	// broadcast address
      	){
      		// TODO: different setup for broadcast and unicast (if necessary)
      		sipr_incoming.type = 0;	// interesting frame: mark "valid flag"
      	} else {
      		sipr_incoming.type = SIPR_PAK_RESERVED;	// neglect frame
      	}
     	break;
    case 4:
    	// don't touch if already decided to neglect
    	// what if not a valid type ? will be handled later
    	if( sipr_incoming.type != SIPR_PAK_RESERVED ){
            sipr_incoming.type = c;
    	}
    	break;
    case 5:
    	if( c != CANID_SIPR ){
    		// oops, mark neglect
    		sipr_incoming.type = SIPR_PAK_RESERVED;
    	}
    	break;
    case 6: // byte 0 in the CAN data, SIPR sequence
    	sipr_incoming.seq = c;
    	break;
    default:
    	if ( idx < sipr_incoming.size + 6 ){
        	// accept and store only MAX_INC_SIZE chars
        	if ( (sipr_incoming.idx - 7) < MAX_INC_SIZE ){
          		sipr_incoming.data[ idx - 7 ] = c;
        	}
      	} else {
    		if ( idx == (sipr_incoming.size + 7)) { // CRC HI byte and check the whole CRC
        		if ((c == HI(sipr_incoming.CRC)) && (sipr_incoming_recCRCl == LO(sipr_incoming.CRC))) {
        			// CRC OK => valid packet
          			sipr_process_packet();
          			return;
        		}
      		}

	      	// either (idx == sipr_incoming.size + 6) (which is CRC LO byte) or we just don't care
    	    sipr_incoming_recCRCl = c;
      		return;
    	}
	}  // switch

  	/* update the CRC */
	sipr_incoming.CRC = _crc_ccitt_update(sipr_incoming.CRC, c);	
}

